#![allow(dead_code)]
use ed25519_dalek::{Signer, SigningKey, VerifyingKey};
use once_cell::sync::Lazy;
use rand::rngs::OsRng;
use std::collections::HashMap;
use std::sync::Mutex;
use std::thread;
use std::time::Duration;
use std::time::SystemTime;
use tabled::{Table, Tabled};

pub enum SigningError {
    CorruptedKey,
    KeyNotFound,
}

impl std::fmt::Display for SigningError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let error_message = match self {
            SigningError::CorruptedKey => "The public key is corrupted",
            SigningError::KeyNotFound => "Did not find the public key",
        };
        write!(f, "{}", error_message)
    }
}

impl From<hex::FromHexError> for SigningError {
    fn from(_: hex::FromHexError) -> Self {
        Self::CorruptedKey
    }
}

impl From<ed25519_dalek::ed25519::Error> for SigningError {
    fn from(_: ed25519_dalek::ed25519::Error) -> Self {
        Self::CorruptedKey
    }
}

impl From<std::array::TryFromSliceError> for SigningError {
    fn from(_: std::array::TryFromSliceError) -> Self {
        Self::CorruptedKey
    }
}

#[derive(Clone)]
pub struct NodeInfo {
    pub pubkey: VerifyingKey,
    pub updated_at: SystemTime,
}

static NODES: Lazy<Mutex<HashMap<String, NodeInfo>>> = Lazy::new(|| Mutex::new(HashMap::new()));

static KEYS: Lazy<Mutex<HashMap<VerifyingKey, SigningKey>>> =
    Lazy::new(|| Mutex::new(HashMap::new()));

pub fn add_key(pubkey: VerifyingKey, privkey: SigningKey) {
    let mut keys = KEYS.lock().unwrap();
    keys.insert(pubkey, privkey);
}

pub fn remove_key(pubkey: &VerifyingKey) {
    let mut keys = KEYS.lock().unwrap();
    keys.remove(pubkey);
}

fn get_privkey(pubkey: &VerifyingKey) -> Option<SigningKey> {
    let keys = KEYS.lock().unwrap();
    keys.get(pubkey).cloned()
}

pub fn sign_message_with_key(pubkey: &str, message: &str) -> Result<String, SigningError> {
    // Parse the hex string into a VerifyingKey
    let key_bytes = hex::decode(pubkey)?;
    let pubkey = VerifyingKey::from_bytes(&key_bytes.as_slice().try_into()?)?;

    // Lock the hashmap and try to get the SigningKey
    let key_store = KEYS.lock().unwrap();
    let signing_key = match key_store.get(&pubkey) {
        Some(k) => k,
        None => return Err(SigningError::KeyNotFound),
    };

    // TODO: check if to_bytes returns the signature in a format that people can verify from bash
    let signature = hex::encode(signing_key.sign(message.as_bytes()).to_bytes());

    Ok(signature)
}

pub fn add_node(ip: String, info: NodeInfo) {
    let mut nodes = NODES.lock().unwrap();
    nodes.insert(ip, info);
}

pub fn remove_node(ip: &str) {
    let mut nodes = NODES.lock().unwrap();
    nodes.remove(ip);
}

pub fn get_pubkey(ip: &str) -> Option<NodeInfo> {
    let nodes = NODES.lock().unwrap();
    nodes.get(ip).cloned()
}

pub fn get_nodes_as_html_tabe() -> String {
    #[derive(Tabled)]
    struct OutputRow {
        ip: String,
        pubkey: String,
        age: u64,
    }
    let mut output = vec![];
    for (ip, node_info) in NODES.lock().unwrap().iter() {
        let ip = ip.clone();
        let pubkey = hex::encode(node_info.pubkey.as_bytes());
        let age = std::time::SystemTime::now()
            .duration_since(node_info.updated_at)
            .unwrap_or(Duration::ZERO)
            .as_secs();
        output.push(OutputRow { ip, pubkey, age });
    }
    Table::new(output).to_string()
}

pub fn cycle_keys() {
    thread::spawn(|| {
        let mut csprng = OsRng;
        loop {
            // TODO: save old private key to disk using SGX Sealing
            let privkey = ed25519_dalek::SigningKey::generate(&mut csprng);
            add_node(
                "localhost".to_string(),
                NodeInfo {
                    pubkey: privkey.verifying_key(),
                    updated_at: std::time::SystemTime::now(),
                },
            );
            add_key(privkey.verifying_key(), privkey);
            thread::sleep(Duration::from_secs(60));
        }
    });
}
